/*R: current memory requirements for GPU part:
gpu_chirps ->32k*4*2*16=4M
ocl_global_buf* ->32k*4*2*32=8M

If unroll by data chunk will be used, then GPU mem requirements will be:
gpu_chirps + N_of_unrolls*(ocl_global_buf1+ocl_global_buf2)=4M+N*16M
for mem size of 128M 7 unrolls are possible.
*/

//#pragma OPENCL EXTENSION cl_amd_printf : enable

#define COMBINED_DECHIRP_KERNEL 1 //R:To use (1) or not to use (0) Frizz's compute-intencive dechirping kernel
#define NO_BESTS 1 //Will govern best pulses search inside find single pulse kernel
__kernel __attribute__((work_group_size_hint(1, 1, 1)))
void set_mem_kernel_cl(__global float* data){
	uint tid=get_global_id(0);
	data[tid]=0.0f;
}
__kernel 
void set_guard_elements(__global float* data, const int offset){
	uint tid=get_global_id(0);
	data[offset+tid]=0.0f;
}
__kernel void splitter_bits_to_float_range_kernel(__global uint4* gpu_raw,__global float4* gpu_data,const uint raw_offset){
//R: I assume that first bit represents data[0].re, second - data[0].im and so on.
//R: Each workitem takes 4 unsigned ints from bit stream, that is 4*4*8 bits or 4*4*8/2 data elements
//R: that is, 64 data elements per work item. We need 32k*DATA_CHUNK_UNROLL elements total =>
//R: there are plenty workitems to load (512*7 in current version) all processing elements of GPU.
//R: raw_offset contains offset into raw array considering raw as array of uint4 entities
//R: (that is, each element contains 4*4*8 bits)
	uint tid= get_global_id(0);
	uint dchunk=get_global_id(1);
	float4 data;
//R: we need 512 work items per data array
	int raw_offset_fin=raw_offset+tid+512/2*dchunk;
	if(raw_offset_fin>=(8*1024*1024>>(2+2))) return;
	uint4 bits=gpu_raw[raw_offset_fin];//R:data chunks overlapped by 50% hence 512/2
	uint data_offset=dchunk*(32768/2)+tid*32;
	for(uint i=0;i<8;i++,data_offset++){
		data.x=2.0f*(bits.x & 1)-1.0f;bits.x>>=1;
		data.y=2.0f*(bits.x & 1)-1.0f;bits.x>>=1;
		data.z=2.0f*(bits.x & 1)-1.0f;bits.x>>=1;
		data.w=2.0f*(bits.x & 1)-1.0f;bits.x>>=1;
		gpu_data[data_offset]=data;
	}
	for(uint i=0;i<8;i++,data_offset++){
		data.x=2.0f*(bits.y & 1)-1.0f;bits.y>>=1;
		data.y=2.0f*(bits.y & 1)-1.0f;bits.y>>=1;
		data.z=2.0f*(bits.y & 1)-1.0f;bits.y>>=1;
		data.w=2.0f*(bits.y & 1)-1.0f;bits.y>>=1;
		gpu_data[data_offset]=data;
	}
	for(uint i=0;i<8;i++,data_offset++){
		data.x=2.0f*(bits.z & 1)-1.0f;bits.z>>=1;
		data.y=2.0f*(bits.z & 1)-1.0f;bits.z>>=1;
		data.z=2.0f*(bits.z & 1)-1.0f;bits.z>>=1;
		data.w=2.0f*(bits.z & 1)-1.0f;bits.z>>=1;
		gpu_data[data_offset]=data;
	}
	for(uint i=0;i<8;i++,data_offset++){
		data.x=2.0f*(bits.w & 1)-1.0f;bits.w>>=1;
		data.y=2.0f*(bits.w & 1)-1.0f;bits.w>>=1;
		data.z=2.0f*(bits.w & 1)-1.0f;bits.w>>=1;
		data.w=2.0f*(bits.w & 1)-1.0f;bits.w>>=1;
		gpu_data[data_offset]=data;
	}
}


 // start of combined dechirp kernel if/else-swith
#if COMBINED_DECHIRP_KERNEL	// must be enabled by hand - not currently set on OpenCL compiler

// dechirping done entirely on GPU

#define sigma 1310.7200f
#define rcp_2SigmaSqr 0.000000291038304567337040f
#define normalize_val 1.05424785025598620f

#if USE_OPENCL_INTEL
//R: Intel SDK includes <stdin> silently that contains Pi definition
#else
#define M_PI 3.14159265358979323846264338f
#endif

#define FFT_SIZE 32768

inline float4 calc_chirp(int j, int i, float dm_start) 
{
	float4 result;
	
	float phase_const = M_PI*(dm_start + (float)i);
	
	float g;
	float freq;
	float multiplier;	
	float phase;

	if(	j < (FFT_SIZE/2) ) {	
	
		float float_j = (float)j;
		g = float_j + 0.5f;
	
		multiplier = native_sqrt(1.0f - native_exp(-(g * g * rcp_2SigmaSqr))) *normalize_val;
	
		freq = float_j/32768.0f;
	
		phase = phase_const * freq * freq * (1.0f / (0.00176f *freq +1.0f));

		result.x = native_cos(phase) * multiplier;
		result.y = native_sin(phase) * multiplier;

		float_j = float_j +1.0f;
		g = float_j + 0.5f;
		multiplier = native_sqrt(1.0f - native_exp(-(g * g * rcp_2SigmaSqr))) *normalize_val;
		
		freq = float_j/32768.0f;
		
		phase = phase_const * freq * freq * (1.0f / (0.00176f *freq +1.0f));

		result.z = native_cos(phase) * multiplier;	
		result.w = native_sin(phase) * multiplier;	
		
	} else 
	{
	
		j = j -16384;

		g = (float)(16383 -j) + 0.5f;
	
		multiplier = native_sqrt(1.0f - native_exp(-(g * g * rcp_2SigmaSqr))) *normalize_val;
	
		freq = (float)(j + 16384)/32768.0f -1.0f;
	
		phase = phase_const * freq * freq * (1.0f / (0.00176f *freq +1.0f));

		result.x = native_cos(phase) * multiplier;
		result.y = native_sin(phase) * multiplier;

		j = j +1;
		g = (float)(16383 -j) + 0.5f;
		multiplier = native_sqrt(1.0f - native_exp(-(g * g * rcp_2SigmaSqr))) *normalize_val;	
		
		freq = (float)(j + 16384)/32768.0f -1.0f;
		
		phase = phase_const * freq * freq * (1.0f / (0.00176f *freq +1.0f));

		result.z = native_cos(phase) * multiplier;	
		result.w = native_sin(phase) * multiplier;	
		
	}
	return result;
}

__kernel void dechirp_range1_kernel ( __global float4* gpu_data ,
										__global float4* gpu_dechirped,
										__const float dm_start)
{

	uint tid = get_global_id(0);
	uint dchunk=get_global_id(1);
	//R: each work item will process 2 complex data elements and write 2*2*16 new complex elements into dechirped array
	
	float4 data=gpu_data[tid+dchunk*(FFT_SIZE/2)];
	float4 cur_chirp;
	float4 cur_dechirp;
	float xx, yy, zz, ww, yx, xy, wz, zw;
	
	for(uint i=0;i<16;i++){//R: can be optimized via mad instruction probably
		
		cur_chirp = calc_chirp(tid *2, i, dm_start);
		
		//negative sign
		xx = data.x*cur_chirp.x;
		yy = data.y*cur_chirp.y;
		yx = data.y*cur_chirp.x;
		xy = data.x*cur_chirp.y;
		zz = data.z*cur_chirp.z;
		ww = data.w*cur_chirp.w;
		wz = data.w*cur_chirp.z;
		zw = data.z*cur_chirp.w;
#if APV505
		cur_dechirp.x = xx - yy;
		cur_dechirp.y = yx + xy;
		cur_dechirp.z = zz - ww;
		cur_dechirp.w = wz + zw;
#else
		cur_dechirp.x = xx + yy;
		cur_dechirp.y = yx - xy;
		cur_dechirp.z = zz + ww;
		cur_dechirp.w = wz - zw;
#endif
		gpu_dechirped[32*(FFT_SIZE/2)*dchunk+(2*i+0)*(FFT_SIZE/2)+tid]=cur_dechirp;

		//positive sign
#if APV505
		cur_dechirp.x = xx + yy;
		cur_dechirp.y = yx - xy;
		cur_dechirp.z = zz + ww;
		cur_dechirp.w = wz - zw;
#else
		cur_dechirp.x = xx - yy;
		cur_dechirp.y = yx + xy;
		cur_dechirp.z = zz - ww;
		cur_dechirp.w = wz + zw;
#endif
		gpu_dechirped[32*(FFT_SIZE/2)*dchunk+(2*i+1)*(FFT_SIZE/2)+tid]=cur_dechirp;

		}
}

#else	// dechirping via look up table

//R: dechirps whole dm from initial single gpu_data array
//data packed in float4 a:
//a.s0<->data[0].re;a.s1<->data[0].im;a.s1<->data[1].re and so on
//in dechirped minus sign goes first, next is positive sign, next is negative for next DM.
__kernel void dechirp_range1_kernel(__global float4* gpu_data ,
                             __global float4* gpu_chirps,
                             __global float4* gpu_dechirped)
{
	uint tid = get_global_id(0);
	uint dchunk=get_global_id(1);
	//R: each work item will process 2 complex data elements and write 2*2*16 new complex elements into dechirped array
	float4 data=gpu_data[tid+dchunk*(32768/2)];
	float4 cur_chirp;
	float4 cur_dechirp;
	for(uint i=0;i<16;i++){//R: can be optimized via mad instruction probably
		cur_chirp=gpu_chirps[i*(32768/2)+tid];
		//negative sign
#if APV505
		cur_dechirp.x = data.x*cur_chirp.x - data.y*cur_chirp.y;
		cur_dechirp.y = data.y*cur_chirp.x + data.x*cur_chirp.y;
		cur_dechirp.z = data.z*cur_chirp.z - data.w*cur_chirp.w;
		cur_dechirp.w = data.w*cur_chirp.z + data.z*cur_chirp.w;
#else
		cur_dechirp.x = data.x*cur_chirp.x + data.y*cur_chirp.y;
		cur_dechirp.y = data.y*cur_chirp.x - data.x*cur_chirp.y;
		cur_dechirp.z = data.z*cur_chirp.z + data.w*cur_chirp.w;
		cur_dechirp.w = data.w*cur_chirp.z - data.z*cur_chirp.w;
#endif
		gpu_dechirped[32*(32768/2)*dchunk+(2*i+0)*(32768/2)+tid]=cur_dechirp;
		//positive sign
#if APV505
		cur_dechirp.x = data.x*cur_chirp.x + data.y*cur_chirp.y;
		cur_dechirp.y = data.y*cur_chirp.x - data.x*cur_chirp.y;
		cur_dechirp.z = data.z*cur_chirp.z + data.w*cur_chirp.w;
		cur_dechirp.w = data.w*cur_chirp.z - data.z*cur_chirp.w;
#else
		cur_dechirp.x = data.x*cur_chirp.x - data.y*cur_chirp.y;
		cur_dechirp.y = data.y*cur_chirp.x + data.x*cur_chirp.y;
		cur_dechirp.z = data.z*cur_chirp.z - data.w*cur_chirp.w;
		cur_dechirp.w = data.w*cur_chirp.z + data.z*cur_chirp.w;
#endif
		gpu_dechirped[32*(32768/2)*dchunk+(2*i+1)*(32768/2)+tid]=cur_dechirp;
	}
}

#endif // end of combined dechirp kernel if/else-swith


__kernel void half_temp_range1_kernel(__global float8* gpu_dechirped, __global float4* gpu_power,const float ravg_pow){
	uint tid = get_global_id(0);
	uint dchunk=get_global_id(1);
	uint i=get_global_id(2);
	
	float8 temp;
	float4 power;

	temp = gpu_dechirped[ (dchunk<<18) + (i<<13) + tid ];

	power.xy =  mad( temp.s02, temp.s02, (temp.s13 * temp.s13) ) * ravg_pow;
		
	power.zw =  mad( temp.s46 ,temp.s46, (temp.s57 * temp.s57) ) * ravg_pow;
		
	gpu_power[ (dchunk<<17) + (i<<12) + tid ] = power;
	
}
#if 0
__kernel void PC_single_pulse_kernel_FFA_update_vectorised(__global float4* gpu_power,
									   __constant float4* thresh, __global uint* results,
									   const int num_dchunks,
									   const int need_small_FFA_data,const int need_large_FFA_data,
									   __global float4* small_neg,__global float4* small_pos,
									   __global float4* large_neg,__global float4* large_pos)
{
	//R: each work-item will handle separate data_chunk/dm/sign power array through all needed coadds
	//R: to avoid host to device data transfers (very costly for current implementation)
	//R: thresholds and best power values will be maintained separately for CPU and GPU
	//R: thresholds are constant, but best pulses will need update.
	//R: need_*_FFA_data are non-zero (if needed) offsets into gpu_power buffer
	//R: we store small FFA buffers (negative first, then positive) right after coadds buffer
	//R: (coadds buffer take same size as initial power array thought it actually slightly smaller, just to simplify
	//R: buffer size calculations for now) then larger FFA buffers go, first negative one, next positive, to match with
	//R: signs order for GPU dechirping.
	uint tid=get_global_id(0);
	uint dchunk=get_global_id(1);
	uint len4 = 32768 >> 3;
	uint4 coadd_offsets=(uint4)(0);
	uint4 coadd_begins=(uint4)(0);
	uint next_begin,next_offset;
	float4 t=thresh[0];//thresholds for first 4 coadds
	float4 tt0=(float4)(t.x);//R:will keep current threshld for vector operations
	float4 tt1=(float4)(t.y);//R:will keep current threshld for vector operations
	float4 tt2=(float4)(t.z);//R:will keep current threshld for vector operations
	float4 tt3=(float4)(t.w);//R:will keep current threshld for vector operations
	uint result=0;//R: will contain info about location of found pulses and best signals
	int4 pulse0=(int4)(0);//R: result of comparison
	int4 pulse1=(int4)(0);//R: result of comparison
	int4 pulse2=(int4)(0);//R: result of comparison
	int4 pulse3=(int4)(0);//R: result of comparison
	uint4 was_pulse=(uint4)(0);
	//R: low 16 bits for pulses above threshold, each coadd level coded as enabled bit in corresponding position
	//R: high 16 bits for best pulses, if were, again, enabled bit means found pulse

//R: will do unroll for coadds to save reads from global memory
	float4 d0,d1,d2,d3,d4,d5,d6,d7,d8,d9,d10,d11,d12,d13,d14,d15;
//R: no arrays in registers still (AMD still had to implement it) so store data in separate registers
	coadd_begins.x=0;//R:initial array - no need to write, in gpu_power domain
	coadd_begins.y=num_dchunks*32*(32768>>3);
	coadd_begins.z=coadd_begins.y+num_dchunks*(32768>>4)*32;
	coadd_begins.w=coadd_begins.z+num_dchunks*(32768>>5)*32;
	next_begin=coadd_begins.w+num_dchunks*(32768>>6)*32;

	coadd_offsets.x=coadd_begins.x+(dchunk*32+tid)*(32768>>3);//l==0
	coadd_offsets.y=coadd_begins.y+(dchunk*32+tid)*(32768>>4);
	coadd_offsets.z=coadd_begins.z+(dchunk*32+tid)*(32768>>5);
	coadd_offsets.w=coadd_begins.w+(dchunk*32+tid)*(32768>>6);
	next_offset=next_begin+(dchunk*32+tid)*(32768>>7);
	for(int m=0;m<len4;m+=16){//R:loading data piece of 16*4 samples
		int m1=coadd_offsets.x+m;
			d0=gpu_power[m1];d1=gpu_power[m1+1];d2=gpu_power[m1+2];d3=gpu_power[m1+3];
			d4=gpu_power[m1+4];d5=gpu_power[m1+5];d6=gpu_power[m1+6];d7=gpu_power[m1+7];
			d8=gpu_power[m1+8];d9=gpu_power[m1+9];d10=gpu_power[m1+10];d11=gpu_power[m1+11];
			d12=gpu_power[m1+12];d13=gpu_power[m1+13];d14=gpu_power[m1+14];d15=gpu_power[m1+15];

			pulse0|=(d0>tt0)|(d1>tt0)|(d2>tt0)|(d3>tt0)|(d4>tt0)|(d5>tt0)|(d6>tt0)|(d7>tt0)|
				(d8>tt0)|(d9>tt0)|(d10>tt0)|(d11>tt0)|(d12>tt0)|(d13>tt0)|(d14>tt0)|(d15>tt0);
			//R: now do coadd and save coadded piece in memory
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			d4.xy=d8.xz+d8.yw;d4.zw=d9.xz+d9.yw;d5.xy=d10.xz+d10.yw;d5.zw=d11.xz+d11.yw;
			d6.xy=d12.xz+d12.yw;d6.zw=d13.xz+d13.yw;d7.xy=d14.xz+d14.yw;d7.zw=d15.xz+d15.yw;
			//R: next coadd level ready to check for pulses
			pulse1|=(d0>tt1)|(d1>tt1)|(d2>tt1)|(d3>tt1)|(d4>tt1)|(d5>tt1)|(d6>tt1)|(d7>tt1);
			//R: coadd level 2 starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			pulse2|=(d0>tt2)|(d1>tt2)|(d2>tt2)|(d3>tt2);
			//R: coadd level 3 starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			pulse3|=(d0>tt3)|(d1>tt3);
			//R: now only coadd and store, need to refresh array in registers so memory read required
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;
			m1=(m>>4)+next_offset;
			gpu_power[m1]=d0;
			//R: level 4 coadd saves data for small FFA sometimes
			if(need_small_FFA_data!=-1){//R: after some data chunk we will not store data for small FFA
				if(tid==0){
				//R: Store negative DM sign power data for small FFA
					small_neg[need_small_FFA_data+(m>>4)+dchunk*(32768>>7)]=d0;
				}else if(tid==1){//R: positive DM sign for small FFA
					small_pos[need_small_FFA_data+(m>>4)+dchunk*(32768>>7)]=d0;
				}
			}
	}//R: initial and 3 coadded arrays are checked, now need to store results of check and reload for next levels
	was_pulse.x=1&(pulse0.x|pulse0.y|pulse0.z|pulse0.w);
	was_pulse.y=1&(pulse1.x|pulse1.y|pulse1.z|pulse1.w);
	was_pulse.z=1&(pulse2.x|pulse2.y|pulse2.z|pulse2.w);
	was_pulse.w=1&(pulse3.x|pulse3.y|pulse3.z|pulse3.w);

	result+=(was_pulse.x)+(was_pulse.y<<1)+(was_pulse.z<<2)+(was_pulse.w<<3);
//R: now lets repeat for next few coadd levels and refill registers array from GPU memory
	t=thresh[1];
	tt0=(float4)(t.x);
	tt1=(float4)(t.y);
	tt2=(float4)(t.z);
	tt3=(float4)(t.w);
	was_pulse=(uint4)(0);
	pulse0=(int4)(0);
	pulse1=(int4)(0);
	pulse2=(int4)(0);
	pulse3=(int4)(0);

	len4 = 32768 >>(3+4);//R:4 coadds were done already
	coadd_begins.x=next_begin;//R: will fetch data stored at the end of prev coadd bunch loop
	coadd_begins.y=coadd_begins.x+num_dchunks*(32768>>7)*32;
	coadd_begins.z=coadd_begins.y+num_dchunks*(32768>>8)*32;
	coadd_begins.w=coadd_begins.z+num_dchunks*(32768>>9)*32;
	next_begin=coadd_begins.w+num_dchunks*(32768>>10)*32;
	coadd_offsets.x=coadd_begins.x+(dchunk*32+tid)*(32768>>7);
	coadd_offsets.y=coadd_begins.y+(dchunk*32+tid)*(32768>>8);
	coadd_offsets.z=coadd_begins.z+(dchunk*32+tid)*(32768>>9);
	coadd_offsets.w=coadd_begins.w+(dchunk*32+tid)*(32768>>10);
	next_offset=next_begin+(dchunk*32+tid)*(32768>>11);
//R:now repeating same loop as for first
	for(int m=0;m<len4;m+=16){//R:loading data piece of 16*4 samples
		int m1=m+coadd_offsets.x;
			d0=gpu_power[m1];d1=gpu_power[m1+1];d2=gpu_power[m1+2];d3=gpu_power[m1+3];
			d4=gpu_power[m1+4];d5=gpu_power[m1+5];d6=gpu_power[m1+6];d7=gpu_power[m1+7];
			d8=gpu_power[m1+8];d9=gpu_power[m1+9];d10=gpu_power[m1+10];d11=gpu_power[m1+11];
			d12=gpu_power[m1+12];d13=gpu_power[m1+13];d14=gpu_power[m1+14];d15=gpu_power[m1+15];
			//R: coadd level 4 check
			pulse0|=(d0>tt0)|(d1>tt0)|(d2>tt0)|(d3>tt0)|(d4>tt0)|(d5>tt0)|(d6>tt0)|(d7>tt0)|
				(d8>tt0)|(d9>tt0)|(d10>tt0)|(d11>tt0)|(d12>tt0)|(d13>tt0)|(d14>tt0)|(d15>tt0);
			//R: now do coadd and save coadded piece in memory
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			d4.xy=d8.xz+d8.yw;d4.zw=d9.xz+d9.yw;d5.xy=d10.xz+d10.yw;d5.zw=d11.xz+d11.yw;
			d6.xy=d12.xz+d12.yw;d6.zw=d13.xz+d13.yw;d7.xy=d14.xz+d14.yw;d7.zw=d15.xz+d15.yw;
			//R: coadd level 5 check
			pulse1|=(d0>tt1)|(d1>tt1)|(d2>tt1)|(d3>tt1)|(d4>tt1)|(d5>tt1)|(d6>tt1)|(d7>tt1);
			//R: next coadd level starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			//R: coadd level 6 check
			pulse2|=(d0>tt2)|(d1>tt2)|(d2>tt2)|(d3>tt2);
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			if(need_large_FFA_data!=-1){//R: need to save data for large FFA only at beginning of large DM chunk
				if(tid==0){
				//R: Store negative DM sign power data for large FFA
					large_neg[need_large_FFA_data+(m>>3)+dchunk*(32768>>10)]=d0;
					large_neg[need_large_FFA_data+(m>>3)+dchunk*(32768>>10)+1]=d1;
				}else if(tid==1){//R: positive DM sign for large FFA
					large_pos[need_large_FFA_data+(m>>3)+dchunk*(32768>>10)]=d0;
					large_pos[need_large_FFA_data+(m>>3)+dchunk*(32768>>10)+1]=d1;
				}
			}

			//R: coadd level 7 check
			pulse3|=(d0>tt3)|(d1>tt3);
			//R: now only coadd and store, need to refresh array in registers so memory read required
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;
			m1=(m>>4)+next_offset;
			gpu_power[m1]=d0;//R: coadd level 8 store
	}
	was_pulse.x=1&(pulse0.x|pulse0.y|pulse0.z|pulse0.w);
	was_pulse.y=1&(pulse1.x|pulse1.y|pulse1.z|pulse1.w);
	was_pulse.z=1&(pulse2.x|pulse2.y|pulse2.z|pulse2.w);
	was_pulse.w=1&(pulse3.x|pulse3.y|pulse3.z|pulse3.w);
	result+=(was_pulse.x<<4)+(was_pulse.y<<5)+(was_pulse.z<<6)+(was_pulse.w<<7);
//R: now lets repeat for next few coadd levels and refill registers array from GPU memory
	t=thresh[2];
	tt0=(float4)(t.x);
	tt1=(float4)(t.y);
	tt2=(float4)(t.z);
	tt3=(float4)(t.w);
	was_pulse=(uint4)(0);
	pulse0=(int4)(0);
	pulse1=(int4)(0);
	pulse2=(int4)(0);
	pulse3=(int4)(0);
	coadd_begins.x=next_begin;//R: will fetch data stored at the end of prev coadd bunch loop
	coadd_begins.y=coadd_begins.x+num_dchunks*(32768>>11)*32;
	coadd_begins.z=coadd_begins.y+num_dchunks*(32768>>12)*32;
	coadd_begins.w=coadd_begins.z+num_dchunks*(32768>>13)*32;
	next_begin=coadd_begins.w+num_dchunks*(32768>>14)*32;
	coadd_offsets.x=coadd_begins.x+(dchunk*32+tid)*(32768>>11);
	coadd_offsets.y=coadd_begins.y+(dchunk*32+tid)*(32768>>12);
	coadd_offsets.z=coadd_begins.z+(dchunk*32+tid)*(32768>>13);
	coadd_offsets.w=coadd_begins.w+(dchunk*32+tid)*(32768>>14);
	next_offset=next_begin+(dchunk*32+tid)*(32768>>15);
//R: 32768>>11==16 so we have only 16 float4 elements, no loop by m needed.
		int m1=coadd_offsets.x;
			d0=gpu_power[m1];d1=gpu_power[m1+1];d2=gpu_power[m1+2];d3=gpu_power[m1+3];
			d4=gpu_power[m1+4];d5=gpu_power[m1+5];d6=gpu_power[m1+6];d7=gpu_power[m1+7];
			d8=gpu_power[m1+8];d9=gpu_power[m1+9];d10=gpu_power[m1+10];d11=gpu_power[m1+11];
			d12=gpu_power[m1+12];d13=gpu_power[m1+13];d14=gpu_power[m1+14];d15=gpu_power[m1+15];
			//R: coadd level 8 check
			pulse0|=(d0>tt0)|(d1>tt0)|(d2>tt0)|(d3>tt0)|(d4>tt0)|(d5>tt0)|(d6>tt0)|(d7>tt0)|
				(d8>tt0)|(d9>tt0)|(d10>tt0)|(d11>tt0)|(d12>tt0)|(d13>tt0)|(d14>tt0)|(d15>tt0);
			//R: now do coadd and save coadded piece in memory
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			d4.xy=d8.xz+d8.yw;d4.zw=d9.xz+d9.yw;d5.xy=d10.xz+d10.yw;d5.zw=d11.xz+d11.yw;
			d6.xy=d12.xz+d12.yw;d6.zw=d13.xz+d13.yw;d7.xy=d14.xz+d14.yw;d7.zw=d15.xz+d15.yw;
			//R: coadd level 9 check,last needed coadd
			pulse1|=(d0>tt1)|(d1>tt1)|(d2>tt1)|(d3>tt1)|(d4>tt1)|(d5>tt1)|(d6>tt1)|(d7>tt1);
	was_pulse.x=1&(pulse0.x|pulse0.y|pulse0.z|pulse0.w);
	was_pulse.y=1&(pulse1.x|pulse1.y|pulse1.z|pulse1.w);
	result+=(was_pulse.x<<8)+(was_pulse.y<<9);
//R: last we do is to write result of pulse finding
	results[32*dchunk+tid]=result;
}

#endif
//R: single pulse is rare event. That is, buffer readback is rare too. So, no need to care about what part of buffer
//should be read back to host memory (and current implementation will read full buffer anyway).
//This allows to save on flags (single flag yes/no needed instead of detailed info where was pulse.
//Also, this allows to split kernel on many just as I did with pulse finding in MB. If one of parts finds pulse
//and corresponding FFA updates not needed, other parts can be skipped completely.
//Splitting kernel will expose greater parallelism and will do much more effective memory latence hiding

//reduce0 will update only small FFA buffers, higher coadds go in next kernel calls
#if 1

__kernel __attribute__((reqd_work_group_size(1, 1, 64)))
void PC_single_pulse_kernel_FFA_update_reduce0(__global float4* gpu_power,
									   __constant float4* thresh, __global uint* results,
									   const int num_dchunks,
									   const int need_small_FFA_data,//const int need_large_FFA_data,
									   __global float4* small_neg,__global float4* small_pos//,
									   //__global float4* large_neg,__global float4* large_pos
									   )
{
	//R: each work-item will handle separate data_chunk/dm/sign power array through all needed coadds
	//R: to avoid host to device data transfers (very costly for current implementation)
	//R: thresholds and best power values will be maintained separately for CPU and GPU
	//R: thresholds are constant, but best pulses will need update.
	//R: need_*_FFA_data are non-zero (if needed) offsets into gpu_power buffer
	//R: we store small FFA buffers (negative first, then positive) right after coadds buffer
	//R: (coadds buffer take same size as initial power array thought it actually slightly smaller, just to simplify
	//R: buffer size calculations for now) then larger FFA buffers go, first negative one, next positive, to match with
	//R: signs order for GPU dechirping.
	uint tid=get_global_id(0);
	uint dchunk=get_global_id(1);
	int mchunk=get_global_id(2);
//	uint len4 = 32768 >> 3;
//	uint4 coadd_offsets=(uint4)(0);
//	uint4 coadd_begins=(uint4)(0);
	uint next_begin,next_offset;
	float4 t=thresh[0];//thresholds for first 4 coadds
	float4 tt0=(float4)(t.x);//R:will keep current threshold for vector operations
	float4 tt1=(float4)(t.y);//R:will keep current threshold for vector operations
	float4 tt2=(float4)(t.z);//R:will keep current threshold for vector operations
	float4 tt3=(float4)(t.w);//R:will keep current threshold for vector operations
	//uint result=0;
	//int4 pulse0=(int4)(0);//R: result of comparison
	//int4 pulse1=(int4)(0);//R: result of comparison
	//int4 pulse2=(int4)(0);//R: result of comparison
	//int4 pulse3=(int4)(0);//R: result of comparison
	int4 was_pulse=(int4)(0);////R: simple yes/no flag: was single pulse found somewhere or not;
	//vector for vectorized ops
	float4 d0,d1,d2,d3,d4,d5,d6,d7,d8,d9,d10,d11,d12,d13,d14,d15;
//R: no arrays in registers still (AMD still had to implement it) so store data in separate registers
	//R: Coadded values storage address arithmetics now greatly simplified
	//R: only 1 coadd level stored per kernel call so max 2 coadds need to be addressed
	next_begin=num_dchunks*32*(32768>>3);//R: area just after initial power array
	next_offset=next_begin+(dchunk*32+tid)*(32768>>7);//R: don't account for mchunk offset

	//coadd_begins.x=0;//R:initial array - no need to write, in gpu_power domain
	//coadd_begins.y=num_dchunks*32*(32768>>3);
	//coadd_begins.z=coadd_begins.y+num_dchunks*(32768>>4)*32;
	//coadd_begins.w=coadd_begins.z+num_dchunks*(32768>>5)*32;
	//next_begin=coadd_begins.w+num_dchunks*(32768>>6)*32;

	//coadd_offsets.x=coadd_begins.x+(dchunk*32+tid)*(32768>>3);//l==0
	//coadd_offsets.y=coadd_begins.y+(dchunk*32+tid)*(32768>>4);
	//coadd_offsets.z=coadd_begins.z+(dchunk*32+tid)*(32768>>5);
	//coadd_offsets.w=coadd_begins.w+(dchunk*32+tid)*(32768>>6);
	//next_offset=next_begin+(dchunk*32+tid)*(32768>>7);
	//for(int m=0;m<len4;m+=16)
	{//R:loading data piece of 16*4 samples
		int m1=(dchunk*32+tid)*(32768>>3)+(mchunk<<4);//R:reading corresponding mchunk (16*4 samples here) from power array - coadd level 0
			d0=gpu_power[m1];d1=gpu_power[m1+1];d2=gpu_power[m1+2];d3=gpu_power[m1+3];
			d4=gpu_power[m1+4];d5=gpu_power[m1+5];d6=gpu_power[m1+6];d7=gpu_power[m1+7];
			d8=gpu_power[m1+8];d9=gpu_power[m1+9];d10=gpu_power[m1+10];d11=gpu_power[m1+11];
			d12=gpu_power[m1+12];d13=gpu_power[m1+13];d14=gpu_power[m1+14];d15=gpu_power[m1+15];

			was_pulse|=(d0>tt0)|(d1>tt0)|(d2>tt0)|(d3>tt0)|(d4>tt0)|(d5>tt0)|(d6>tt0)|(d7>tt0)|
				(d8>tt0)|(d9>tt0)|(d10>tt0)|(d11>tt0)|(d12>tt0)|(d13>tt0)|(d14>tt0)|(d15>tt0);
			//R: now do coadd and save coadded piece in registers
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			d4.xy=d8.xz+d8.yw;d4.zw=d9.xz+d9.yw;d5.xy=d10.xz+d10.yw;d5.zw=d11.xz+d11.yw;
			d6.xy=d12.xz+d12.yw;d6.zw=d13.xz+d13.yw;d7.xy=d14.xz+d14.yw;d7.zw=d15.xz+d15.yw;
			//R: next coadd level ready to check for pulses
			was_pulse|=(d0>tt1)|(d1>tt1)|(d2>tt1)|(d3>tt1)|(d4>tt1)|(d5>tt1)|(d6>tt1)|(d7>tt1);
			//R: coadd level 2 starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			was_pulse|=(d0>tt2)|(d1>tt2)|(d2>tt2)|(d3>tt2);
			//R: coadd level 3 starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			was_pulse|=(d0>tt3)|(d1>tt3);
			//R: now only coadd and store, need to refresh array in registers so memory read required
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;
			m1=mchunk+next_offset;
			gpu_power[m1]=d0;
			//R: level 4 coadd saves data for small FFA sometimes
			if(need_small_FFA_data!=-1){//R: after some data chunk we will not store data for small FFA
				if(tid==0){
				//R: Store negative DM sign power data for small FFA
					small_neg[need_small_FFA_data+mchunk+dchunk*(32768>>7)]=d0;
				}else if(tid==1){//R: positive DM sign for small FFA
					small_pos[need_small_FFA_data+mchunk+dchunk*(32768>>7)]=d0;
				}
			}
	}
	was_pulse.x=was_pulse.x||was_pulse.y||was_pulse.z||was_pulse.w;//result reduction
	if(was_pulse.x) results[0]=1;
}
__kernel __attribute__((reqd_work_group_size(32, 1, 2)))
void PC_single_pulse_kernel_FFA_update_reduce1(__global float4* gpu_power,
									   __constant float4* thresh, __global uint* results,
									   const int num_dchunks,
									   //const int need_small_FFA_data,
									   const int need_large_FFA_data,
									   //__global float4* small_neg,__global float4* small_pos//,
									   __global float4* large_neg,__global float4* large_pos
									   )
{
	uint tid=get_global_id(0);
	uint dchunk=get_global_id(1);
	uint mchunk=get_global_id(2);
//R: now lets repeat for next few coadd levels and refill registers array from GPU memory
	float4 t=thresh[1];
	float4 tt0=(float4)(t.x);//R:will keep current threshold for vector operations
	float4 tt1=(float4)(t.y);//R:will keep current threshold for vector operations
	float4 tt2=(float4)(t.z);//R:will keep current threshold for vector operations
	float4 tt3=(float4)(t.w);//R:will keep current threshold for vector operations
	t=thresh[2];//for last 2 levels
	int4 was_pulse=(int4)(0);////R: simple yes/no flag: was single pulse found somewhere or not;
	//vector for vectorized ops
	float4 d0,d1,d2,d3,d4,d5,d6,d7,d8,d9,d10,d11,d12,d13,d14,d15;
	int initial_begin=num_dchunks*32*(32768>>3);
	int initial_offset=initial_begin+(dchunk*32+tid)*(32768>>7);//R:we reading already coadded data here, for particular dchunk
	//int next_offset=initial_begin+num_dchunks*(32768>>7)*32+(dchunk*32+tid)*(32768>>11);
	//len4 = 32768 >>(3+4);//R:4 coadds were done already

	//coadd_begins.x=next_begin;//R: will fetch data stored at the end of prev coadd bunch loop
	//coadd_begins.y=coadd_begins.x+num_dchunks*(32768>>7)*32;
	//coadd_begins.z=coadd_begins.y+num_dchunks*(32768>>8)*32;
	//coadd_begins.w=coadd_begins.z+num_dchunks*(32768>>9)*32;
	//next_begin=coadd_begins.w+num_dchunks*(32768>>10)*32;
	//coadd_offsets.x=coadd_begins.x+(dchunk*32+tid)*(32768>>7);
	//coadd_offsets.y=coadd_begins.y+(dchunk*32+tid)*(32768>>8);
	//coadd_offsets.z=coadd_begins.z+(dchunk*32+tid)*(32768>>9);
	//coadd_offsets.w=coadd_begins.w+(dchunk*32+tid)*(32768>>10);
	//next_offset=next_begin+(dchunk*32+tid)*(32768>>11);
//R:now repeating same loop as for first
	//for(int m=0;m<len4;m+=16)
	{//R:loading data piece of 16*4 samples
		int m1=(mchunk<<4)+initial_offset;
			d0=gpu_power[m1];d1=gpu_power[m1+1];d2=gpu_power[m1+2];d3=gpu_power[m1+3];
			d4=gpu_power[m1+4];d5=gpu_power[m1+5];d6=gpu_power[m1+6];d7=gpu_power[m1+7];
			d8=gpu_power[m1+8];d9=gpu_power[m1+9];d10=gpu_power[m1+10];d11=gpu_power[m1+11];
			d12=gpu_power[m1+12];d13=gpu_power[m1+13];d14=gpu_power[m1+14];d15=gpu_power[m1+15];
			//R: coadd level 4 check
			was_pulse|=(d0>tt0)|(d1>tt0)|(d2>tt0)|(d3>tt0)|(d4>tt0)|(d5>tt0)|(d6>tt0)|(d7>tt0)|
				(d8>tt0)|(d9>tt0)|(d10>tt0)|(d11>tt0)|(d12>tt0)|(d13>tt0)|(d14>tt0)|(d15>tt0);
			//R: now do coadd and save coadded piece in registers
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			d4.xy=d8.xz+d8.yw;d4.zw=d9.xz+d9.yw;d5.xy=d10.xz+d10.yw;d5.zw=d11.xz+d11.yw;
			d6.xy=d12.xz+d12.yw;d6.zw=d13.xz+d13.yw;d7.xy=d14.xz+d14.yw;d7.zw=d15.xz+d15.yw;
			//R: coadd level 5 check
			was_pulse|=(d0>tt1)|(d1>tt1)|(d2>tt1)|(d3>tt1)|(d4>tt1)|(d5>tt1)|(d6>tt1)|(d7>tt1);
			//R: next coadd level starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			//R: coadd level 6 check
			was_pulse|=(d0>tt2)|(d1>tt2)|(d2>tt2)|(d3>tt2);
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			if(need_large_FFA_data!=-1){//R: need to save data for large FFA only at beginning of large DM chunk
				if(tid==0){
				//R: Store negative DM sign power data for large FFA
					large_neg[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)]=d0;
					large_neg[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)+1]=d1;
				}else if(tid==1){//R: positive DM sign for large FFA
					large_pos[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)]=d0;
					large_pos[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)+1]=d1;
				}
			}

			//R: coadd level 7 check
			was_pulse|=(d0>tt3)|(d1>tt3);
			was_pulse.x|=was_pulse.y|was_pulse.z|was_pulse.w;//result reduction
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;
			//R: coadd level 8 check
			was_pulse.x|=(d0.x>t.x)|(d0.y>t.x)|(d0.z>t.x)|(d0.w>t.x);
			//R: now do coadd and save coadded piece in registers
			d0.xy=d0.xz+d0.yw;
			//R: coadd level 9 check,last needed coadd
			was_pulse.x|=(d0.x>t.y)|(d0.y>t.y);
	}
	if(was_pulse.x) results[0]=1;
}
#else
//R: difference from prev ersion: smaller registers array (possible more active blocks per compute unit)
//and full use of coadds even if writes will be not float4 (they will be write combined anyway)
//Another possible advantage - smaller memory fetch stride - less or no memory banks/channels conflicts
__kernel void PC_single_pulse_kernel_FFA_update_reduce0(__global const float4* gpu_power,__global float* coadded,
									   __constant float4* thresh, __global uint* results,
									   const int num_dchunks,
									   const int need_small_FFA_data,//const int need_large_FFA_data,
									   __global float2* small_neg,__global float2* small_pos//,
									   //__global float4* large_neg,__global float4* large_pos
									   )
{
	//R: each work-item will handle separate data_chunk/dm/sign power array through all needed coadds
	//R: to avoid host to device data transfers (very costly for current implementation)
	//R: thresholds and best power values will be maintained separately for CPU and GPU
	//R: thresholds are constant, but best pulses will need update.
	//R: need_*_FFA_data are non-zero (if needed) offsets into gpu_power buffer
	//R: we store small FFA buffers (negative first, then positive) right after coadds buffer
	//R: (coadds buffer take same size as initial power array thought it actually slightly smaller, just to simplify
	//R: buffer size calculations for now) then larger FFA buffers go, first negative one, next positive, to match with
	//R: signs order for GPU dechirping.
	uint tid=get_global_id(0);
	uint dchunk=get_global_id(1);
	uint mchunk=get_global_id(2);
//	uint len4 = 32768 >> 3;
//	uint4 coadd_offsets=(uint4)(0);
//	uint4 coadd_begins=(uint4)(0);
	uint next_begin,next_offset;
	float4 t=thresh[0];//thresholds for first 4 coadds
	float4 tt0=(float4)(t.x);//R:will keep current threshold for vector operations
	float4 tt1=(float4)(t.y);//R:will keep current threshold for vector operations
	float4 tt2=(float4)(t.z);//R:will keep current threshold for vector operations
	float4 tt3=(float4)(t.w);//R:will keep current threshold for vector operations
	//uint result=0;
	//int4 pulse0=(int4)(0);//R: result of comparison
	//int4 pulse1=(int4)(0);//R: result of comparison
	//int4 pulse2=(int4)(0);//R: result of comparison
	//int4 pulse3=(int4)(0);//R: result of comparison
	int4 was_pulse=(int4)(0);////R: simple yes/no flag: was single pulse found somewhere or not;
	//vector for vectorized ops
	float4 d0,d1,d2,d3,d4,d5,d6,d7;
//R: no arrays in registers still (AMD still had to implement it) so store data in separate registers
	//R: Coadded values storage address arithmetics now greatly simplified
	//R: only 1 coadd level stored per kernel call so max 2 coadds need to be addressed
	next_begin=num_dchunks*32*(32768>>3);//R: area just after initial power array
	next_offset=next_begin+(dchunk*32+tid)*(32768>>8);//R: don't account for mchunk offset

	//coadd_begins.x=0;//R:initial array - no need to write, in gpu_power domain
	//coadd_begins.y=num_dchunks*32*(32768>>3);
	//coadd_begins.z=coadd_begins.y+num_dchunks*(32768>>4)*32;
	//coadd_begins.w=coadd_begins.z+num_dchunks*(32768>>5)*32;
	//next_begin=coadd_begins.w+num_dchunks*(32768>>6)*32;

	//coadd_offsets.x=coadd_begins.x+(dchunk*32+tid)*(32768>>3);//l==0
	//coadd_offsets.y=coadd_begins.y+(dchunk*32+tid)*(32768>>4);
	//coadd_offsets.z=coadd_begins.z+(dchunk*32+tid)*(32768>>5);
	//coadd_offsets.w=coadd_begins.w+(dchunk*32+tid)*(32768>>6);
	//next_offset=next_begin+(dchunk*32+tid)*(32768>>7);
	//for(int m=0;m<len4;m+=16)
	{//R:loading data piece of 16*4 samples
		int m1=mchunk<<3;//R:reading corresponding mchunk (8*4 samples here) from power array - coadd level 0
			d0=gpu_power[m1];d1=gpu_power[m1+1];d2=gpu_power[m1+2];d3=gpu_power[m1+3];
			d4=gpu_power[m1+4];d5=gpu_power[m1+5];d6=gpu_power[m1+6];d7=gpu_power[m1+7];

			was_pulse|=(d0>tt0)|(d1>tt0)|(d2>tt0)|(d3>tt0)|(d4>tt0)|(d5>tt0)|(d6>tt0)|(d7>tt0);
			//R: now do coadd and save coadded piece in registers
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			//R: next coadd level ready to check for pulses
			was_pulse|=(d0>tt1)|(d1>tt1)|(d2>tt1)|(d3>tt1);
			//R: coadd level 2 starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			was_pulse|=(d0>tt2)|(d1>tt2);
			//R: coadd level 3 starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;
			was_pulse|=(d0>tt3);
			d0.xy=d0.xz+d0.yw;
			t=thresh[1];
			was_pulse.xy|=was_pulse.zw;
			was_pulse.xy|=(d0.xy>t.xx);
			//R: level 4 coadd saves data for small FFA sometimes
			if(need_small_FFA_data!=-1){//R: after some data chunk we will not store data for small FFA
				if(tid==0){//R: array of float2 now
				//R: Store negative DM sign power data for small FFA
					small_neg[(need_small_FFA_data<<1)+mchunk+dchunk*(32768>>7)]=d0.xy;
				}else if(tid==1){//R: positive DM sign for small FFA
					small_pos[(need_small_FFA_data<<1)+mchunk+dchunk*(32768>>7)]=d0.xy;
				}
			}
			d0.x+=d0.y;
			was_pulse.x|=was_pulse.y|(d0.x>t.y);
			m1=mchunk+next_offset;
			coadded[m1]=d0.x;

	}
	if(was_pulse.x) results[0]=1;
}
#endif
/*
__kernel void PC_single_pulse_kernel_FFA_update_reduce2(__global float4* gpu_power,
									   __constant float4* thresh, __global uint* results,
									   const int num_dchunks
									   )
{
	uint tid=get_global_id(0);
	uint dchunk=get_global_id(1);
	uint mchunk=get_global_id(2);
//R: now lets repeat for next few coadd levels and refill registers array from GPU memory
//R: only 2 levels remain so to improve parallelization and reduce register pressure let's use smaller registers array
	float4 t=thresh[2];
	//tt2=(float4)(t.z);
	//tt3=(float4)(t.w);
	int was_pulse=(0);////R: simple yes/no flag: was single pulse found somewhere or not;
	//vector for vectorized ops
	float4 d0;
//	pulse0=(int4)(0);
//	pulse1=(int4)(0);
//	pulse2=(int4)(0);
//	pulse3=(int4)(0);
	int initial_offset=num_dchunks*32*(32768>>3)+num_dchunks*(32768>>7)*32+(dchunk*32+tid)*(32768>>11);//R: will fetch data stored at the end of prev coadd bunch loop
	d0=gpu_power[initial_offset];
			//R: coadd level 8 check
			was_pulse=(d0.x>t.x)|(d0.y>t.x)|(d0.z>t.x)|(d0.w>t.x);
			//R: now do coadd and save coadded piece in registers
			d0.xy=d0.xz+d0.yw;
			//R: coadd level 9 check,last needed coadd
			was_pulse|=(d0.x>t.y)|(d0.y>t.y);
	if(was_pulse) results[0]=1;
}
*/
//R: block of support functions for df64 DP emulation
float __fmul_rn(float a, float b){return a*b;}
float __fadd_rn(float a, float b){return a+b;}
float __float2int_rn(float a){return rint(a);}
float __int2float_rn(int a){return (float)a;}
float2 split(float a)
{
	const float split= 4097;
	float t = __fmul_rn(a,split);
	float2 r;
	r.x = t-(t-a);
	r.y = a - r.x;
	return r;
}
 float2 twoProd(float a, float b)
{
	float2 r;
	r.x = __fmul_rn(a,b);
	float2 aS = split(a);
	float2 bS = split(b);
	r.y = ( ( __fmul_rn(aS.x,bS.x) - r.x )
				+ __fmul_rn(aS.x,bS.y) + __fmul_rn(aS.y,bS.x) )
				+ __fmul_rn(aS.y,bS.y);
	return r;
}

float2 quickTwoSum(float a, float b)
{
	float2 r;
	r.x = a + b;
	r.y = b - (r.x-a);
	return r;
}

float2 df64_mult(float2 a, float2 b)
{
	float ah=a.x,al=a.y,bh=b.x,bl=b.y;
	float2 p; 
	p = twoProd(ah,bh);
	p.y += ah*bl;
	p.y += al*bh;
	p = quickTwoSum(p.x,p.y);
	return p;
}

#if 0
__kernel void GPU_fetch_array_kernel_df64_cl(__global float* src, __constant float2* periods, __global int* f_int, __global float4* dest,
		const uint offset, const uint stride){
	uint threadID=get_global_id(0)+offset;
	uint j=get_global_id(1);
	int k=0;
	int l=0;
	float4 acc=(float4)0.f;
	int n_per=f_int[threadID];
	float2 period=periods[threadID];
	for(k=0;k<n_per;k++){
		float2 jj=split(__int2float_rn(k));
		float2 tmp=df64_mult(period,jj);
		float int_part=trunc(tmp.x);
		tmp.x-=int_part;
		tmp.x+=tmp.y;
			//l=offsets[k*stride+threadID];
			l=((int)int_part)+((int)rint(tmp.x));
			l+=(4*j);//R: index to data array computed
			acc.x+=src[l];
			acc.y+=src[l+1];
			acc.z+=src[l+2];
		 	acc.w+=src[l+3];
#if 0
	if(l >= 65534){
		printf("threadID=%d; j=%d; n_per=%d\n",threadID,j,n_per);
	 	printf("l=%d; k=%d; array value:(%.4g, %.4g, %.4g, %.4g)\n",l,k,src[l],src[l+1],src[l+2],src[l+3]);
	}
#endif
	}
	dest[threadID*480+j]=acc;
}
#elif 1
__kernel void GPU_fetch_array_kernel_cl(__global float* src, __global int* offsets, __global int* f_int, __global float4* dest,
		const uint offset, const uint stride){
	uint threadID=get_global_id(0)+offset;
	uint j=get_global_id(1);
	int k=0;
	int l=0;
	float4 acc=(float4)0.f;
	int n_per=f_int[threadID];
	for(k=0;k<n_per;k++){
			l=offsets[k*stride+threadID];
			l+=(4*j);//R: index to data array computed
			acc.x+=src[l];
			acc.y+=src[l+1];
			acc.z+=src[l+2];
		 	acc.w+=src[l+3];
#if 0
	if(l >= 65534){
		printf("threadID=%d; j=%d; n_per=%d\n",threadID,j,n_per);
	 	printf("l=%d; k=%d; array value:(%.4g, %.4g, %.4g, %.4g)\n",l,k,src[l],src[l+1],src[l+2],src[l+3]);
	}
#endif

	}
	dest[threadID*480+j]=acc;
}
#else
//uje: Will fill at end of blocks upto grid size with "zeroes"
__kernel void GPU_fetch_array_fill_kernel_cl(__global float* src, __global int* offsets, __global int* f_int, __global float4* dest,
    const uint offset, const uint max_threads_per_block, const uint filler) {
  uint threadID=get_global_id(0)+offset;
  uint j=get_global_id(1);
  int k=0;
  int l=0;
  float4 acc=(float4)0.f;
  if(filler==0) { //uje: should be ok except for last blocks
    int n_per=f_int[threadID];
    for(k=0;
        k<n_per;
        k++) {
      l=offsets[k*max_threads_per_block+threadID];
      l+=(4*j); //R: index to data array computed
      acc.x+=src[l];
      acc.y+=src[l+1];
      acc.z+=src[l+2];
      acc.w+=src[l+3];
    }
    dest[threadID*480+j]=acc;
  }
  else {
    int n_per=f_int[threadID];
    for(k=0;
        k<n_per;
        k++) {
      l=offsets[k*max_threads_per_block+threadID];
      l+=(4*j); //R: index to data array computed
      acc.x+=src[l];
      acc.y+=src[l+1];
      acc.z+=src[l+2];
      acc.w+=src[l+3];
    }
    if(threadID<(max_threads_per_block-filler)) {
      dest[threadID*480+j]=acc;
    }
    //uje: if there are less than max_threads_per_block threads in use fill the rest with "zeroes"
    else {
      dest[threadID*480+j]=(float4)0.f;
    }
  }
}
#endif
__kernel void GPU_compare_with_threshold_kernel_cl(const int size, __global float4* input, __global float* thresholds,
												   const int level, __global int* output, const uint stride)
{
	int threadID=get_global_id(0);
	float threshold=thresholds[level+threadID*12];
	float4 p;
	int i=size-1;
	//int was_signal=0;
	int bin=99999;
	//int ln=(len[threadID]+3)>>2;
	for(;i>=0;i--){
		p=input[i+threadID*stride];
		if(p.w>=threshold){
			//was_signal++;
			bin=4*i+3;
		}
		if(p.z>=threshold){
			//was_signal++;
			bin=4*i+2;
		}
		if(p.y>=threshold){
			//was_signal++;
			bin=4*i+1;
		}
		if(p.x>=threshold){
			//was_signal++;
			bin=4*i;
		}
	}
	//output.x=was_signal;
	output[threadID]=bin;
}
__kernel void GPU_PC_compare_with_threshold_kernel_cl(const int size, __global float4* input, __global float* thresholds,
												   const int level, __global int* output, const uint stride,__global int* array_size)
{
	int threadID=get_global_id(0);
	float threshold=thresholds[level+threadID*12];
	float4 p;
	int i=size-1;
	//int was_signal=0;
	int bin=99999;
	//int ln=(len[threadID]+3)>>2;
	for(;i>=0;i--){
		p=input[i+threadID*stride];
		if(p.w>=threshold){
			//was_signal++;
			bin=4*i+3;
		}
		if(p.z>=threshold){
			//was_signal++;
			bin=4*i+2;
		}
		if(p.y>=threshold){
			//was_signal++;
			bin=4*i+1;
		}
		if(p.x>=threshold){
			//was_signal++;
			bin=4*i;
		}
	}
	//output.x=was_signal;
	if(bin<array_size[threadID]){
		output[0]=1+threadID;
	}
}
#if 1
__kernel void GPU_coadd_kernel_cl(__global float4* src, __global float4* dest, const uint stride)
{
	int threadID=get_global_id(0);
	int i=get_global_id(1);
	uint new_stride;
	float4 o;
	float4 i2;
	float4 i21;
		i2=src[2*i+threadID*stride];
		i21=src[2*i+1+threadID*stride];
		o.xy=i2.xz+i2.yw;
		o.zw=i21.xz+i21.yw;
		if(stride>120){
			new_stride=(stride>>1);
		}else{
			new_stride=stride;
		}
#if 0
		if(threadID==0 && i==0) printf("stride=%d; new_stride=%d\n",stride,new_stride);
#endif
		dest[i+threadID*(new_stride)]=o;
}
#else
//uje: 1D, if 2D operation does not work correct e.g. on linux with HD4670
__kernel void GPU_coadd_1D_kernel_cl(const uint size, __global float4* src, __global float4* dest, const uint stride) {
  int threadID = get_global_id(0);
  int i = 0;
  uint new_stride;
  float4 o;
  float4 i2;
  float4 i21;
  for(;
      i<4*size;
      i++) {
    i2 = src[2*i+threadID*stride];
    i21 = src[2*i+1+threadID*stride];
    o.xy = i2.xz+i2.yw;
    o.zw = i21.xz+i21.yw;
    if(stride>120) {
      new_stride = (stride>>1);
    } else {
      new_stride = stride;
    }
    dest[i+threadID*new_stride] = o;
  }
}
#endif
#if 0 //USE_OPENCL_NV && !_GNU_SOURCE //R: non-vectorized version
__kernel void GPU_coadd_with_stride_kernel_cl(__global float* src,__global int* stride,__global float* dest,const uint stride1)
{
	int threadID=get_global_id(0);
	int i=get_global_id(1);
	float o=src[i+threadID*(stride1*4)];
	int st=stride[threadID];
	o+=src[i+st+4*stride1*threadID];
	dest[threadID*240*4+i]=o;
}
#else
__kernel void GPU_coadd_with_stride_kernel_cl(__global float4* src,__global int* stride,__global float4* dest,const uint stride1)
{
	int threadID=get_global_id(0);
	int i=get_global_id(1);
	float4 o=src[i+threadID*stride1];
	int st=stride[threadID];
	int offset=st>>2;
	int s=st-4*(offset);
	float4 i1=src[i+offset+stride1*threadID];
	float4 i2=src[i+offset+1+stride1*threadID];
	//if(i >= per_int[treadID]) return;
	if(s==0){
		o+=i1;
	}else if(s==1){
		o.xyz+=i1.yzw;
		o.w+=i2.x;
	}else if(s==2){
		o.xy+=i1.zw;
		o.zw+=i2.xy;
	}else if(s==3){
		o.x+=i1.w;
		o.yzw+=i2.xyz;
	}
	dest[threadID*240+i]=o;
}
#endif
#if 0
//uje: not using float4 with this 1D kernel
__kernel void GPU_coadd_with_stride_1D_kernel_cl(const uint size,__global float* src,__global int* stride,__global float* dest,const uint stride1) {
  int threadID=get_global_id(0);
  int i=0;
  float o;
  int st=stride[threadID];
  for(;
      i<size;
      i++) {
    o=src[i+4*stride1*threadID];
    o+=src[i+st+4*stride1*threadID];
    dest[threadID*240*4+i]=o;
  }
}
#endif
__kernel void GPU_change_array_sizes_kernel_cl(__global int* array_sizes){
	array_sizes[get_global_id(0)] >>=1;
}
__kernel void GPU_change_array_sizes_per_int_kernel_cl(__global int* array_sizes,__global int* overlap){
	const uint tid=get_global_id(0);
	int v=array_sizes[tid];
	overlap[tid] =  (v >> 1);
    array_sizes[tid] = ((v + 1)  >> 1);
}
#if defined(USE_OPENCL_NV) && defined(OCL_FERMI) 
#pragma OPENCL EXTENSION cl_khr_fp64 : enable  
__kernel void PopulateFetchOffsets_kernel_DP_cl(__global double* periods,__global int* offsets){
	int tid=get_global_id(0);
	int j=get_global_id(1);
	offsets[tid+j*get_global_size(0)]=(int)(periods[tid]*j + 0.5);
}
#endif
#if USE_OCL_DP //R: take advantage of DP card if host app supports this path
#if defined(cl_khr_fp64)   //R:double precision chirp here
#pragma OPENCL EXTENSION cl_khr_fp64 : enable  
#endif
#if defined(cl_amd_fp64)
#pragma OPENCL EXTENSION cl_amd_fp64 : enable  
#endif
//R: f_int can be in constant space if threads number will be less than 16k always 
__kernel void PopulateTresholdTable_kernel9t(int dm_chunk, __global int* f_int,
										 double thresh_x_intercept,
										 double thresh_x_slope,__global float* thresh_tbl){
	int threadID=get_global_id(0);
	int ttidx_pre =gert_global_id(1);
	//int nfb= ((int)freq[threadID]);  
	int nfb=f_int[threadID];
	int ncb= (nfb*dm_chunk)<<ttidx_pre;
	double thresh_x;
	nfb=nfb<<ttidx_pre;
          thresh_x = thresh_x_intercept + thresh_x_slope * (double)nfb;
          thresh_tbl = (float)(ncb + (int)(thresh_x *(double) sqrt(2.0f * (float) ncb)));
}
#else //R: FP-only implementation if possible
__kernel void PopulateTresholdTable_kernel9t_df64_cl(int dm_chunk, __global int* f_int,
										 float2 thresh_x_intercept,//R:in split format
										 float2 thresh_x_slope,//R: in split format
										 __global float* thresh_tbl){
	int threadID=get_global_id(0);
	int ttidx_pre =get_global_id(1);
	//int nfb= ((int)freq[threadID]);  
	int nfb=f_int[threadID];
	int ncb= (nfb*dm_chunk)<<ttidx_pre;
	float2 thresh_x;//R:in split format
	nfb=nfb<<ttidx_pre;
	int tmp=ncb;
    thresh_x = thresh_x_intercept + df64_mult(thresh_x_slope, split(__int2float_rn(nfb)));
	float2 tmp_df64=df64_mult(thresh_x,split(sqrt(2.0f * (float) ncb)));
	tmp+=(int)(tmp_df64.x+tmp_df64.y);
	thresh_tbl[threadID*get_global_size(1)+ttidx_pre] = (float)tmp;
}
__kernel void PopulateFetchOffsets_kernel_df64_cl(__global float2* periods,__global int* offsets){
	int tid=get_global_id(0);
	int j=get_global_id(1);
	//float2 half=split(0.5f);
	float2 jj=split(__int2float_rn(j));
	float2 tmp=df64_mult(periods[tid],jj);
	//for(unsigned j=0;j<n_per[tid];j++)
	float int_part=trunc(tmp.x);
	tmp.x-=int_part;
	tmp.x+=tmp.y;
	offsets[tid+j*get_global_size(0)]=((int)int_part)+((int)rint(tmp.x));
}
#endif